local CLAN_OPCODE = 23
local Profile = false

local IInventorySlotStyles = {
  [InventorySlotHead] = "HeadSlot",
  [InventorySlotNeck] = "NeckSlot",
  [InventorySlotBack] = "BackSlot",
  [InventorySlotBody] = "BodySlot",
  [InventorySlotRight] = "RightSlot",
  [InventorySlotLeft] = "LeftSlot",
  [InventorySlotLeg] = "LegSlot",
  [InventorySlotFeet] = "FeetSlot",
  [InventorySlotFinger] = "FingerSlot",
  [InventorySlotAmmo] = "AmmoSlot",

}

local inspectWindow = nil

function init()
  connect(
    g_game,
    {
      onGameStart = create,
      onGameEnd = destroy
    }
  )

  ProtocolGame.registerExtendedOpcode(CLAN_OPCODE, SendClanInfo)

  if g_game.isOnline() then
    create()
  end
end

function terminate()
  disconnect(
    g_game,
    {
      onGameStart = create,
      onGameEnd = destroy
    }
  )

  ProtocolGame.unregisterExtendedOpcode(CLAN_OPCODE)

  destroy()
end

function create()
  inspectWindow = g_ui.displayUI("inspect")
  inspectWindow:hide()
  g_game.getProtocolGame():sendExtendedOpcode(CLAN_OPCODE, json.encode{protocol = "item", slot = slot, item = item, name = name})
  g_game.getProtocolGame():sendExtendedOpcode(CLAN_OPCODE, json.encode{protocol = "stats", stats = stats, showItems = showItems})
end

function SendClanInfo(protocol, opcode, buffer)
    local receive = json.decode(buffer)
    local protocol = receive.protocol	
	if protocol == "item" then	
		addItem(receive.slot, receive.item, receive.name)
	elseif protocol == "stats" then	
		addStats(receive.stats)
		sendShowItems = receive.showItems
	elseif protocol == "inspect" then
		addItem(receive.slot, receive.item, receive.name)	
		addStats(receive.stats)
	end
		
	if playerName == g_game.getCharacterName() then
		inspectWindow:getChildById("Account"):setVisible(true)
		inspectWindow:recursiveGetChildById("ShowItems"):setVisible(true)
		else
		inspectWindow:getChildById("Account"):setVisible(false)
		inspectWindow:recursiveGetChildById("ShowItems"):setVisible(false)
	end
end

function inspect(creature)
   g_game.getProtocolGame():sendExtendedOpcode(CLAN_OPCODE, json.encode{protocol = "inspect", playerName = creature})
   playerName = creature
 toggle()
end

function addStats(data)
  -- Atualizar os elementos de texto e barra de progresso
  local function updateTextAndValue(widgetId, text, value, maxValue)
    local w = inspectWindow:recursiveGetChildById(widgetId)
    if w then
      w:setText(text)
      if value and maxValue then
        w:setValue(value, 0, maxValue)
      end
    end
  end

  -- Atualizar estat�sticas b�sicas
  for key, value in pairs(data) do
    local w = inspectWindow:recursiveGetChildById(key)
    if w then
      w:setText(type(value) == "number" and comma_value(value) or value)
    end
  end

  updateTextAndValue("healthBar", data.health .. "/" .. data.maxHealth, data.health, data.maxHealth)
  updateTextAndValue("manaBar", data.mana .. "/" .. data.maxMana, data.mana, data.maxMana)

  -- Atualizar habilidades
  for i, skill in ipairs(data.skills) do
    local w = inspectWindow:recursiveGetChildById("skill" .. i)
    if w then
      local valueWidget = w:getChildById("value")
      valueWidget:setText(skill.total)
      if skill.bonus > 0 then
        valueWidget:setColor("#008b00")
        w:setTooltip("Skill: "..skill.total - skill.bonus .. " \nBonus: " .. skill.bonus)
      else
        valueWidget:setColor("#bbbbbb")
        w:removeTooltip()
      end
      local percentWidget = w:getChildById("percent")
      percentWidget:setPercent(math.floor(skill.percent))
      percentWidget:setTooltip(tr("%s percent to go", 100 - skill.percent))
    end
  end

  -- Atualizar outras estat�sticas
  local stats = {
    {id = "textMagic", value = data.magic},
    {id = "stat1Text", value = data.criticalChange.. "%"},
    {id = "stat2Text", value = data.healthBonus.. "%"},
    {id = "stat3Text", value = data.manaBonus.. "%"},
    {id = "stat4Text", value = data.critDamage.. "%"},
    {id = "stat5Text", value = data.magicBonus.. "%"}, --Healing Bonus
    {id = "stat6Text", value = data.damage.. "%"},
    {id = "stat7Text", value = data.defense.. "%"},
    {id = "stat8Text", value = data.lifeLeech.. "%"},
    {id = "stat9Text", value = data.manaLeech.. "%"},
    {id = "stat10Text", value = data.luck},
	  {id = "stat11Text", value = data.extraexp.. "%"},
	  {id = "stat12Text", value = data.attackSpeed.. "%"},
	  {id = "stat13Text", value = data.healthRegen.. "%"},
	  {id = "stat14Text", value = data.manaRegen.. "%"},
  }

  for _, stat in ipairs(stats) do
    local statWidget = inspectWindow:recursiveGetChildById(stat.id)
    if statWidget then
      statWidget:setText(stat.value)
    end
  end

  -- Atualizar outfit
  local outfitCreatureBox = inspectWindow:recursiveGetChildById("outfitInspectBox")
  if outfitCreatureBox then
    outfitCreatureBox:setOutfit({type = data.outfit})
		if Profile then 
		   HTTP.downloadImage(data.imagem, function(path, err)
			if err then g_logger.warning("HTTP error: " .. err .. " - " .. data.imagem) return end
			outfitCreatureBox:setImageSource(path)
			end)
		end
    end
  
  inspectWindow:setText(data.name.." Infos")

  
local ShowItems = inspectWindow:recursiveGetChildById("ShowItems")
local isChecked = false  

if sendShowItems == true then
  isChecked = false
else
  isChecked = true
end

ShowItems:setChecked(isChecked)

local function showItemsButtonClick()
  isChecked = not isChecked 
  ShowItems:setChecked(isChecked)  
  g_game.getProtocolGame():sendExtendedOpcode(CLAN_OPCODE, json.encode{protocol = "showItem", isChecked = isChecked})
end

ShowItems.onClick = showItemsButtonClick
end


function addStats22(data)
  for key, value in pairs(data) do
    local w = inspectWindow:recursiveGetChildById(key)
    if w then
      w:setText(type(value) == "number" and comma_value(value) or value)
    end
  end
  local healthBar = inspectWindow:recursiveGetChildById("healthBar")
  healthBar:setText(data.health .. "/" .. data.maxHealth)
  healthBar:setValue(data.health, 0, data.maxHealth)
  local manaBar = inspectWindow:recursiveGetChildById("manaBar")
  manaBar:setText(data.mana .. "/" .. data.maxMana)
  manaBar:setValue(data.mana, 0, data.maxMana)

  for i = 1, #data.skills do
    local skill = data.skills[i]
    local w = inspectWindow:recursiveGetChildById("skill" .. i)
    if w then
      local valueWidget = w:getChildById("value")
      valueWidget:setText(skill.total)
      if skill.bonus > 0 then
        valueWidget:setColor("#008b00")
        w:setTooltip(skill.total - skill.bonus .. " +" .. skill.bonus)
      else
        valueWidget:setColor("#bbbbbb")
        w:removeTooltip()
      end
      local percentWidget = w:getChildById("percent")
      percentWidget:setPercent(math.floor(skill.percent))
      percentWidget:setTooltip(tr("%s percent to go", 100 - skill.percent))
    end
  end
    
  local MagicLevel = inspectWindow:recursiveGetChildById("textMagic")
  MagicLevel:setText(data.magic)

  local Stat1 = inspectWindow:recursiveGetChildById("stat1Text")
  Stat1:setText(data.criticalChange)
  
    local Stat2 = inspectWindow:recursiveGetChildById("stat2Text")
  Stat2:setText(data.healthBonus)
  
    local Stat3 = inspectWindow:recursiveGetChildById("stat3Text")
  Stat3:setText(data.manaBonus)
  
    local Stat4 = inspectWindow:recursiveGetChildById("stat4Text")
  Stat4:setText(data.skillsBonus)
  
    local Stat5 = inspectWindow:recursiveGetChildById("stat5Text")
  Stat5:setText(data.magicBonus)
  
    local Stat6 = inspectWindow:recursiveGetChildById("stat6Text")
  Stat6:setText(data.damage)
  
    local Stat7 = inspectWindow:recursiveGetChildById("stat7Text")
  Stat7:setText(data.defense)
  
    local Stat8 = inspectWindow:recursiveGetChildById("stat8Text")
  Stat8:setText(data.lifeLeech)
  
    local Stat9 = inspectWindow:recursiveGetChildById("stat9Text")
  Stat9:setText(data.manaLeech)
  
  local outfitCreatureBox = inspectWindow:recursiveGetChildById("outfitInspectBox")
 outfitCreatureBox:setOutfit({type = data.outfit})

  inspectWindow:show()
end

function addItem(slot, item, name)
  local inventoryPanel = inspectWindow:getChildById("inventoryPanel")
  local itemWidget = inventoryPanel:getChildById("slot" .. slot)
  
  if not itemWidget then
    return
  end

  if item then
    itemWidget:setItemId(item)
    itemWidget:setStyle("InventoryItem")
	itemWidget:setTooltip(name)
  else
    itemWidget:setItem(nil)
    itemWidget:setStyle(IInventorySlotStyles[slot])
	itemWidget:setTooltip(nil)
  end
end


function destroy()
  if inspectWindow then
    inspectWindow:destroy()
    inspectWindow = nil
  end
end

function toggle()
    if not g_game.isOnline() then
        return
    end

    if inspectWindow:isVisible() then
        inspectWindow:hide()
    else
        inspectWindow:show()
    end
end

function show()
  if not inspectWindow then
    return
  end  
  inspectWindow:show()
  inspectWindow:raise()
  inspectWindow:focus()
end

function hide()
  if not inspectWindow then
    return
  end
  inspectWindow:hide()
end
